package com.google.gson.util;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.StringReader;

import java.util.List;


/**
 * It's not multi-thread safety.
 * @author haiyangl
 *
 */
public class BufferedReader extends Reader {
    protected static final int DEFAULT_BUFFER_SIZE = 8192;
    protected static final List<char[]> bufferPool = N.newArrayList(100);
    protected char[] cbuff;
    protected int nChars = 0;
    protected int nextChar = 0;
    protected boolean skipLF = false;
    protected Reader in;

    public BufferedReader(String st) {
        this(new StringReader(st));
    }

    public BufferedReader(InputStream is) {
        this(new InputStreamReader(is, N.UTF_8));
    }

    public BufferedReader(Reader reader) {
        in = reader;
        lock = reader;
    }

    /**
     * Reads a single character.
     *
     * @return The character read, as an integer in the range
     *         0 to 65535 (<tt>0x00-0xffff</tt>), or -1 if the
     *         end of the stream has been reached
     * @exception  IOException  If an I/O error occurs
     */
    public int read() throws IOException {
        do {
            if (nextChar >= nChars) {
                fill();

                if (nextChar >= nChars) {
                    return -1;
                }
            }

            if (skipLF) {
                skipLF = false;

                if (cbuff[nextChar] == '\n') {
                    nextChar++;

                    continue;
                }
            }

            return cbuff[nextChar++];
        } while (true);
    }

    /**
     * Reads characters into a portion of an array, reading from the underlying
     * stream if necessary.
     */
    private int read1(char[] cbuf, int off, int len) throws IOException {
        if (nextChar >= nChars) {
            /* If the requested length is at least as large as the buffer, and
               if there is no mark/reset activity, and if line feeds are not
               being skipped, do not bother to copy the characters into the
               local buffer.  In this way buffered streams will cascade
               harmlessly. */
            if ((len >= DEFAULT_BUFFER_SIZE) && !skipLF) {
                return in.read(cbuf, off, len);
            }

            fill();
        }

        if (nextChar >= nChars) {
            return -1;
        }

        if (skipLF) {
            skipLF = false;

            if (cbuff[nextChar] == '\n') {
                nextChar++;

                if (nextChar >= nChars) {
                    fill();
                }

                if (nextChar >= nChars) {
                    return -1;
                }
            }
        }

        int n = Math.min(len, nChars - nextChar);
        System.arraycopy(cbuff, nextChar, cbuf, off, n);
        nextChar += n;

        return n;
    }

    /**
     * Reads characters into a portion of an array.
     *
     * <p> This method implements the general contract of the corresponding
     * <code>{@link Reader#read(char[], int, int) read}</code> method of the
     * <code>{@link Reader}</code> class.  As an additional convenience, it
     * attempts to read as many characters as possible by repeatedly invoking
     * the <code>read</code> method of the underlying stream.  This iterated
     * <code>read</code> continues until one of the following conditions becomes
     * true: <ul>
     *
     *   <li> The specified number of characters have been read,
     *
     *   <li> The <code>read</code> method of the underlying stream returns
     *   <code>-1</code>, indicating end-of-file, or
     *
     *   <li> The <code>ready</code> method of the underlying stream
     *   returns <code>false</code>, indicating that further input requests
     *   would block.
     *
     * </ul> If the first <code>read</code> on the underlying stream returns
     * <code>-1</code> to indicate end-of-file then this method returns
     * <code>-1</code>.  Otherwise this method returns the number of characters
     * actually read.
     *
     * <p> Subclasses of this class are encouraged, but not required, to
     * attempt to read as many characters as possible in the same fashion.
     *
     * <p> Ordinarily this method takes characters from this stream's character
     * buffer, filling it from the underlying stream as necessary.  If,
     * however, the buffer is empty, the mark is not valid, and the requested
     * length is at least as large as the buffer, then this method will read
     * characters directly from the underlying stream into the given array.
     * Thus redundant <code>BufferedReader</code>s will not copy data
     * unnecessarily.
     *
     * @param      cbuf  Destination buffer
     * @param      off   Offset at which to start storing characters
     * @param      len   Maximum number of characters to read
     *
     * @return     The number of characters read, or -1 if the end of the
     *             stream has been reached
     *
     * @exception  IOException  If an I/O error occurs
     */
    public int read(char[] cbuf, int off, int len) throws IOException {
        if ((off < 0) || (off > cbuf.length) || (len < 0) || ((off + len) > cbuf.length) || ((off + len) < 0)) {
            throw new IndexOutOfBoundsException();
        } else if (len == 0) {
            return 0;
        }

        int n = read1(cbuf, off, len);

        if (n <= 0) {
            return n;
        }

        while ((n < len) && in.ready()) {
            int n1 = read1(cbuf, off + n, len - n);

            if (n1 <= 0) {
                break;
            }

            n += n1;
        }

        return n;
    }

    /**
     * Reads a line of text.  A line is considered to be terminated by any one
     * of a line feed ('\n'), a carriage return ('\r'), or a carriage return
     * followed immediately by a linefeed.
     *
     * @param      ignoreLF  If true, the next '\n' will be skipped
     *
     * @return     A String containing the contents of the line, not including
     *             any line-termination characters, or null if the end of the
     *             stream has been reached
     *
     * @see        java.io.LineNumberReader#readLine()
     *
     * @exception  IOException  If an I/O error occurs
     */
    String readLine(boolean ignoreLF) throws IOException {
        StringBuilder sb = null;
        int startChar;

        boolean omitLF = ignoreLF || skipLF;

        do {
            if (nextChar >= nChars) {
                fill();
            }

            if (nextChar >= nChars) { /* EOF */

                if ((sb != null) && (sb.length() > 0)) {
                    return sb.toString();
                } else {
                    return null;
                }
            }

            boolean eol = false;
            char c = 0;
            int i;

            /* Skip a leftover '\n', if necessary */
            if (omitLF && (cbuff[nextChar] == '\n')) {
                nextChar++;
            }

            skipLF = false;
            omitLF = false;

            for (i = nextChar; i < nChars; i++) {
                c = cbuff[i];

                if ((c == '\n') || (c == '\r')) {
                    eol = true;

                    break;
                }
            }

            startChar = nextChar;
            nextChar = i;

            if (eol) {
                String str;

                if (sb == null) {
                    str = new String(cbuff, startChar, i - startChar);
                } else {
                    sb.append(cbuff, startChar, i - startChar);
                    str = sb.toString();
                }

                nextChar++;

                if (c == '\r') {
                    skipLF = true;
                }

                ObjectFactory.recycle(sb);

                return str;
            }

            if (sb == null) {
                sb = ObjectFactory.createStringBuilder();
            }

            sb.append(cbuff, startChar, i - startChar);
        } while (true);
    }

    /**
     * Reads a line of text.  A line is considered to be terminated by any one
     * of a line feed ('\n'), a carriage return ('\r'), or a carriage return
     * followed immediately by a linefeed.
     *
     * @return     A String containing the contents of the line, not including
     *             any line-termination characters, or null if the end of the
     *             stream has been reached
     *
     * @exception  IOException  If an I/O error occurs
     */
    public String readLine() throws IOException {
        return readLine(false);
    }

    /**
     * Skips characters.
     *
     * @param  n  The number of characters to skip
     *
     * @return    The number of characters actually skipped
     *
     * @exception  IllegalArgumentException  If <code>n</code> is negative.
     * @exception  IOException  If an I/O error occurs
     */
    public long skip(long n) throws IOException {
        if (n < 0L) {
            throw new IllegalArgumentException("skip value is negative");
        }

        long r = n;

        while (r > 0) {
            if (nextChar >= nChars) {
                fill();
            }

            if (nextChar >= nChars) { /* EOF */

                break;
            }

            if (skipLF) {
                skipLF = false;

                if (cbuff[nextChar] == '\n') {
                    nextChar++;
                }
            }

            long d = nChars - nextChar;

            if (r <= d) {
                nextChar += r;
                r = 0;

                break;
            } else {
                r -= d;
                nextChar = nChars;
            }
        }

        return n - r;
    }

    /**
     * Tells whether this stream is ready to be read.  A buffered character
     * stream is ready if the buffer is not empty, or if the underlying
     * character stream is ready.
     *
     * @exception  IOException  If an I/O error occurs
     */
    public boolean ready() throws IOException {
        /*
         * If newline needs to be skipped and the next char to be read
         * is a newline character, then just skip it right away.
         */
        if (skipLF) {
            /* Note that in.ready() will return true if and only if the next
             * read on the stream will not block.
             */
            if ((nextChar >= nChars) && in.ready()) {
                fill();
            }

            if (nextChar < nChars) {
                if (cbuff[nextChar] == '\n') {
                    nextChar++;
                }

                skipLF = false;
            }
        }

        return (nextChar < nChars) || in.ready();
    }

    public void close() throws IOException {
        try {
            if (in != null) {
                in.close();
            }
        } finally {
            _reset();
        }
    }

    void setReader(Reader reader) {
        this.in = reader;
        this.lock = reader;
    }

    void _reset() {
        synchronized (lock) {
            recycleBuffer(cbuff);
            cbuff = null;
            nextChar = 0;
            nChars = 0;
            skipLF = false;

            in = null;
        }
    }

    protected void fill() throws IOException {
        if (cbuff == null) {
            cbuff = createBuffer();
        }

        int len = nChars - nextChar;

        if (len > 0) {
            System.arraycopy(cbuff, nextChar, cbuff, 0, len);
        }

        nextChar = 0;
        nChars = len;

        int n = in.read(cbuff, len, cbuff.length - len);

        if (n > 0) {
            nChars += n;
        }
    }

    protected char[] createBuffer() {
        synchronized (bufferPool) {
            return (bufferPool.size() == 0) ? new char[DEFAULT_BUFFER_SIZE] : bufferPool.remove(bufferPool.size() - 1);
        }
    }

    protected void recycleBuffer(char[] cbuff) {
        if ((cbuff != null) && (cbuff.length == DEFAULT_BUFFER_SIZE)) {
            synchronized (bufferPool) {
                bufferPool.add(cbuff);
            }
        }
    }
}
