package com.google.gson.util;

import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.Reader;
import java.io.StringReader;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;


public class ObjectFactory {
    static final int POOLABLE_SIZE = 8192;
    private static final int MAX_POOL_SIZE = 1024;
    private static final int MAX_ARRAY_POOL_SIZE = 320000;
    private static final Queue<List<?>> listPool = N.newArrayBlockingQueue(MAX_POOL_SIZE);
    private static final Queue<Set<?>> setPool = N.newArrayBlockingQueue(MAX_POOL_SIZE);
    private static final Queue<Map<?, ?>> mapPool = N.newArrayBlockingQueue(MAX_POOL_SIZE);
    private static final Queue<LinkedHashMap<?, ?>> linkedHashMapPool = N.newArrayBlockingQueue(MAX_POOL_SIZE);
    private static final Map<Integer, Queue<Object[]>> objectArrayPool = N.asConcurrentHashMap();
    private static final Queue<char[]> charArrayBufferPool = N.newArrayBlockingQueue(MAX_POOL_SIZE);
    private static final Queue<byte[]> byteArrayBufferPool = N.newArrayBlockingQueue(MAX_POOL_SIZE);
    private static final Queue<StringBuilder> stringBuilderPool = N.newArrayBlockingQueue(MAX_POOL_SIZE);
    private static final Queue<ByteArrayOutputStream> byteArrayOutputStreamPool = N.newArrayBlockingQueue(MAX_POOL_SIZE);
    private static final Queue<BufferedWriter> bufferedWriterPool = N.newArrayBlockingQueue(MAX_POOL_SIZE);
    private static final Queue<BufferedReader> bufferedReaderPool = N.newArrayBlockingQueue(MAX_POOL_SIZE);

    @SuppressWarnings({ "unchecked", "rawtypes" })
	public static <T> List<T> createList() {
        List<T> list = (List<T>) listPool.poll();

        return (list == null) ? new ArrayList() : list;
    }

    @SuppressWarnings({ "unchecked", "rawtypes" })
	public static <T> List<T> createList(int initCapacity) {
        List<T> list = (List<T>) listPool.poll();

        return (list == null) ? new ArrayList(initCapacity) : list;
    }

    @SuppressWarnings("unchecked")
	public static <T> Set<T> createSet() {
        Set<T> set = (Set<T>) setPool.poll();

        return (set == null) ? new HashSet<T>() : set;
    }

    @SuppressWarnings("unchecked")
	public static <K, V> Map<K, V> createMap() {
        Map<K, V> map = (Map<K, V>) mapPool.poll();

        return (map == null) ? new HashMap<K, V>() : map;
    }

    @SuppressWarnings("unchecked")
	public static <K, V> Map<K, V> createMap(Object... a) {
        if ((a.length % 2) != 0) {
            throw new IllegalArgumentException(
                "the parameter must be the pairs of property name and value, or an entity class with get/set methods.");
        }

        Map<K, V> map = createMap();

        for (int i = 0; i < a.length; i++) {
            map.put((K) a[i], (V) a[++i]);
        }

        return map;
    }

    @SuppressWarnings("unchecked")
	public static <K, V> LinkedHashMap<K, V> createLinkedHashMap() {
        LinkedHashMap<K, V> linkedHashMap = (LinkedHashMap<K, V>) linkedHashMapPool.poll();

        return (linkedHashMap == null) ? new LinkedHashMap<K, V>() : linkedHashMap;
    }

    @SuppressWarnings("unchecked")
	public static <K, V> LinkedHashMap<K, V> createLinkedHashMap(Object... a) {
        if ((a.length % 2) != 0) {
            throw new IllegalArgumentException(
                "the parameter must be the pairs of property name and value, or an entity class with get/set methods.");
        }

        LinkedHashMap<K, V> map = createLinkedHashMap();

        for (int i = 0; i < a.length; i++) {
            map.put((K) a[i], (V) a[++i]);
        }

        return map;
    }

    public static Object[] createObjectArray() {
        return createObjectArray(POOLABLE_SIZE);
    }

    public static Object[] createObjectArray(int size) {
        Object[] objArray = null;
        Queue<Object[]> arrayQueue = objectArrayPool.get(size);

        if (N.notNullOrEmpty(arrayQueue)) {
            objArray = arrayQueue.poll();
        }

        return (objArray == null) ? new Object[size] : objArray;
    }

    public static char[] createCharArrayBuffer() {
        return createCharArrayBuffer(POOLABLE_SIZE);
    }

    public static char[] createCharArrayBuffer(int capacity) {
        if (capacity > POOLABLE_SIZE) {
            return new char[capacity];
        }

        char[] cbuf = charArrayBufferPool.poll();

        return (cbuf == null) ? new char[POOLABLE_SIZE] : cbuf;
    }

    public static byte[] createByteArrayBuffer() {
        return createByteArrayBuffer(POOLABLE_SIZE);
    }

    public static byte[] createByteArrayBuffer(int capacity) {
        if (capacity > POOLABLE_SIZE) {
            return new byte[capacity];
        }

        byte[] bbuf = byteArrayBufferPool.poll();

        return (bbuf == null) ? new byte[POOLABLE_SIZE] : bbuf;
    }

    public static StringBuilder createStringBuilder() {
        return createStringBuilder(POOLABLE_SIZE);
    }

    public static StringBuilder createStringBuilder(int initCapacity) {
        if (initCapacity > POOLABLE_SIZE) {
            return new StringBuilder(initCapacity);
        }

        StringBuilder sb = stringBuilderPool.poll();

        return (sb == null) ? new StringBuilder(POOLABLE_SIZE) : sb;
    }

    public static ByteArrayOutputStream createByteArrayOutputStream() {
        return createByteArrayOutputStream(POOLABLE_SIZE);
    }

    public static ByteArrayOutputStream createByteArrayOutputStream(int initCapacity) {
        if (initCapacity > POOLABLE_SIZE) {
            return new ByteArrayOutputStream(initCapacity);
        }

        ByteArrayOutputStream os = byteArrayOutputStreamPool.poll();

        return (os == null) ? new ByteArrayOutputStream(POOLABLE_SIZE) : os;
    }

    public static BufferedWriter createBufferedWriter() {
        BufferedWriter bufWriter = bufferedWriterPool.poll();

        if (bufWriter == null) {
            bufWriter = new BufferedWriter(ObjectFactory.createStringBuilder());
        } else {
            if (bufWriter.getStringBuilder() == null) {
                bufWriter.setStringBuilder(ObjectFactory.createStringBuilder());
            }
        }

        bufWriter.setStringBuilderRecycable(true);

        return bufWriter;
    }

    public static BufferedWriter createBufferedWriter(StringBuilder sb) {
        BufferedWriter bufWriter = bufferedWriterPool.poll();

        if (bufWriter == null) {
            bufWriter = new BufferedWriter(sb);
        } else {
            if (bufWriter.isStringBuilderRecycable()) {
                ObjectFactory.recycle(bufWriter.getStringBuilder());
            }

            bufWriter.setStringBuilder(sb);
        }

        bufWriter.setStringBuilderRecycable(false);

        return bufWriter;
    }

    public static BufferedWriter createBufferedWriter(OutputStream os) {
        BufferedWriter bufWriter = bufferedWriterPool.poll();

        if (bufWriter == null) {
            bufWriter = new BufferedWriter(os);
        } else {
            if (bufWriter.isStringBuilderRecycable()) {
                ObjectFactory.recycle(bufWriter.getStringBuilder());
            }

            bufWriter.setWriter(os);
        }

        bufWriter.setStringBuilderRecycable(false);

        return bufWriter;
    }

    public static BufferedWriter createBufferedWriter(Writer writer) {
        if (writer instanceof BufferedWriter) {
            return (BufferedWriter) writer;
        }

        BufferedWriter bufWriter = bufferedWriterPool.poll();

        if (bufWriter == null) {
            bufWriter = new BufferedWriter(writer);
        } else {
            if (bufWriter.isStringBuilderRecycable()) {
                ObjectFactory.recycle(bufWriter.getStringBuilder());
            }

            bufWriter.setWriter(writer);
        }

        bufWriter.setStringBuilderRecycable(false);

        return bufWriter;
    }

    public static BufferedReader createBufferedReader(String st) {
        BufferedReader bufReader = bufferedReaderPool.poll();

        if (bufReader == null) {
            return new BufferedReader(st);
        } else {
            bufReader.setReader(new StringReader(st));

            return bufReader;
        }
    }

    public static BufferedReader createBufferedReader(InputStream is) {
        BufferedReader bufReader = bufferedReaderPool.poll();

        if (bufReader == null) {
            return new BufferedReader(is);
        } else {
            bufReader.setReader(new InputStreamReader(is));

            return bufReader;
        }
    }

    public static BufferedReader createBufferedReader(Reader reader) {
        if (reader instanceof BufferedReader) {
            return (BufferedReader) reader;
        }

        BufferedReader bufReader = bufferedReaderPool.poll();

        if (bufReader == null) {
            return new BufferedReader(reader);
        } else {
            bufReader.setReader(reader);

            return bufReader;
        }
    }

    public static void recycle(List<?> list) {
        if ((list == null) || (list.size() > POOLABLE_SIZE)) {
            return;
        }

        synchronized (listPool) {
            if (listPool.size() < MAX_POOL_SIZE) {
                list.clear();
                listPool.add(list);
            }
        }
    }

    public static void recycle(Set<?> set) {
        if ((set == null) || (set.size() > POOLABLE_SIZE)) {
            return;
        }

        synchronized (setPool) {
            if (setPool.size() < MAX_POOL_SIZE) {
                set.clear();
                setPool.add(set);
            }
        }
    }

    public static void recycle(Map<?, ?> map) {
        if ((map == null) || (map.size() > POOLABLE_SIZE)) {
            return;
        }

        if (map instanceof LinkedHashMap) {
            synchronized (linkedHashMapPool) {
                if (linkedHashMapPool.size() < MAX_POOL_SIZE) {
                    map.clear();
                    linkedHashMapPool.add((LinkedHashMap<?, ?>) map);
                }
            }
        } else {
            synchronized (mapPool) {
                if (mapPool.size() < MAX_POOL_SIZE) {
                    map.clear();
                    mapPool.add(map);
                }
            }
        }
    }

    public static void recycle(Object[] objArray) {
        if ((objArray == null) || (objArray.length > POOLABLE_SIZE)) {
            return;
        }

        synchronized (objectArrayPool) {
            Queue<Object[]> arrayQueue = objectArrayPool.get(objArray.length);

            int poolSize = MAX_ARRAY_POOL_SIZE / objArray.length;

            if (arrayQueue == null) {
                arrayQueue = N.newArrayBlockingQueue(poolSize);
                objectArrayPool.put(objArray.length, arrayQueue);
            }

            if (arrayQueue.size() < poolSize) {
                Arrays.fill(objArray, null);
                arrayQueue.add(objArray);
            }
        }
    }

    public static void recycle(char[] cbuf) {
        if ((cbuf == null) || (cbuf.length > POOLABLE_SIZE)) {
            return;
        }

        synchronized (charArrayBufferPool) {
            if (charArrayBufferPool.size() < MAX_POOL_SIZE) {
                charArrayBufferPool.add(cbuf);
            }
        }
    }

    public static void recycle(byte[] bbuf) {
        if ((bbuf == null) || (bbuf.length > POOLABLE_SIZE)) {
            return;
        }

        synchronized (byteArrayBufferPool) {
            if (byteArrayBufferPool.size() < MAX_POOL_SIZE) {
                byteArrayBufferPool.add(bbuf);
            }
        }
    }

    public static void recycle(StringBuilder sb) {
        if ((sb == null) || (sb.capacity() > POOLABLE_SIZE)) {
            return;
        }

        synchronized (stringBuilderPool) {
            if (stringBuilderPool.size() < MAX_POOL_SIZE) {
                sb.setLength(0);
                stringBuilderPool.add(sb);
            }
        }
    }

    public static void recycle(ByteArrayOutputStream os) {
        if ((os == null) || (os.size() > POOLABLE_SIZE)) {
            return;
        }

        synchronized (byteArrayOutputStreamPool) {
            if (byteArrayOutputStreamPool.size() < MAX_POOL_SIZE) {
                os.reset();
                byteArrayOutputStreamPool.add(os);
            }
        }
    }

    public static void recycle(BufferedWriter writer) {
        if (writer == null) {
            return;
        }

        synchronized (bufferedWriterPool) {
            if (writer.isStringBuilderRecycable()) {
                recycle(writer.getStringBuilder());
            }

            writer._reset();

            if (bufferedWriterPool.size() < MAX_POOL_SIZE) {
                bufferedWriterPool.add(writer);
            }
        }
    }

    public static void recycle(BufferedReader reader) {
        if (reader == null) {
            return;
        }

        synchronized (bufferedWriterPool) {
            reader._reset();

            if (bufferedReaderPool.size() < MAX_POOL_SIZE) {
                bufferedReaderPool.add(reader);
            }
        }
    }
}
